------------------------------------------------------------------------------
--
--  package Base_Types (spec)
--
--  Base types for the ICFP2002 robot controller.
--
------------------------------------------------------------------------------
--  Standard packages:

with Ada.Text_IO;

------------------------------------------------------------------------------

package Base_Types is

   type Robot_ID is new Integer;

   type Gift_ID is new Integer;

   type Gift_Counter is range 0 .. 10_000;

   type Tile_Types is (Plain, Water, Wall, Home);

   type Tile_State (Occupied : Boolean := False) is
      record
         Tile                : Tile_Types;
         Net_Dropped_Gifts   : Gift_Counter;
--       Lost_Gifts          : Natural;
--       Known_Gifts         : array of Gifts;

         case Occupied is
            when True =>
               Robot : Robot_ID;
            when False =>
               null;
         end case;
      end record;

   ---------------------------------------------------------------------------

   type X_Coordinate is range 0 .. 1_001;
   type Y_Coordinate is range 0 .. 1_001;

   type Position_Type is
      record
         X : X_Coordinate;
         Y : Y_Coordinate;
      end record;

   ---------------------------------------------------------------------------

   type Directions is (North, East, South, West);

   procedure Translate (Point     : in out Position_Type;
                        Direction : in     Directions);

   function Translate (Point     : in     Position_Type;
                       Direction : in     Directions) return Position_Type;

   ---------------------------------------------------------------------------

   type Board_Type is array (X_Coordinate range <>,
                             Y_Coordinate range <>) of Tile_State;

   type Actions is (Move, Pick, Drop);

   type Gift_List is array (Gift_Counter range <>) of Gift_ID;

   type Command (Action     : Actions := Drop;
                 Gift_Count : Gift_Counter := 0) is
      record
         case Action is
            when Move =>
               Direction : Directions;
            when Pick | Drop =>
               Gifts     : Gift_List (1 .. Gift_Count);
         end case;
      end record;

   Null_Command : constant Command :=
     (Action     => Drop,
      Gift_Count => 0,
      Gifts      => (others => Gift_ID'First));

   type Weight_Type is new Natural;

   type Gift_Type (Known : Boolean := False) is
      record
         ID : Gift_ID;

         case Known is
            when True =>
               Destination : Position_Type;
               Weight      : Weight_Type;
            when False =>
               null;
         end case;
      end record;

   type Gift_Type_List is array (Gift_Counter range <>) of Gift_Type;

   type Bid_Type is range -1_000_000_000 .. 1_000_000_000;
   subtype Money is Bid_Type range 0 .. Bid_Type'Last;

   type Robot_Configuration is
      record
         ID        : Robot_ID;
         Capacity  : Weight_Type;
         Cash      : Money;
      end record;

   type Robot_Type is
      record
         ID       : Robot_ID;
         Position : Position_Type;
         Gifts    : Gift_Counter;
      end record;

   ---------------------------------------------------------------------------
   --  ID text IO:

   package Robot_ID_Text_IO is
     new Ada.Text_IO.Integer_IO (Num => Robot_ID);

   package Gift_ID_Text_IO is
     new Ada.Text_IO.Integer_IO (Num => Gift_ID);

   ---------------------------------------------------------------------------
   --  Coordinate text IO:

   package X_Coordinate_Text_IO is
     new Ada.Text_IO.Integer_IO (Num => X_Coordinate);

   package Y_Coordinate_Text_IO is
     new Ada.Text_IO.Integer_IO (Num => Y_Coordinate);

   ---------------------------------------------------------------------------
   --  Weight text IO:

   package Weight_Text_IO is
     new Ada.Text_IO.Integer_IO (Num => Weight_Type);

   ---------------------------------------------------------------------------
   --  Money text IO:

   package Money_Text_IO is
     new Ada.Text_IO.Integer_IO (Num => Bid_Type);

   ---------------------------------------------------------------------------

end Base_Types;
