------------------------------------------------------------------------------
--
--  package Landscapes (body)
--
------------------------------------------------------------------------------
--  Standard packages:

with Ada.Integer_Text_IO;
with Ada.Text_IO;

------------------------------------------------------------------------------
--  Other packages:

------------------------------------------------------------------------------

package body Landscapes is

   ---------------------------------------------------------------------------
   --  procedure To_Landscape:

   procedure To_Landscape (Source : in     Base_Types.Board_Type;
                           Target :    out Landscape;
                           Map    : in     Mapping := Default_Map) is

   begin --  To_Landscape
      for X in Target'Range (1) loop
         for Y in Target'Range (2) loop
            if X in Source'Range (1) and Y in Source'Range (2) then
               Target (X, Y) := Map (Source (X, Y).Tile);
            else
               Target (X, Y) := (Interpolation => Ignore);
            end if;
         end loop;
      end loop;
   end To_Landscape;

   ------------------------------------------------------------------------
   --  procedure Adjust:
   --
   --  Sets the height of a point to be the average value of the heights of
   --  the neighbouring points.

   procedure Adjust (Item : in out Landscape;
                     X    : in     Base_Types.X_Coordinate;
                     Y    : in     Base_Types.Y_Coordinate) is

      Midpoint_Weight : constant := 1.0;

      use Base_Types;

      Neighbour : Position_Type;
      Sum       : Float := 0.0;
      Count     : Natural := 0;

   begin --  Adjust
      if X in Item'Range (1) and Y in Item'Range (2) then
         if Item (X, Y).Interpolation = Flexible then
            for Direction in Directions loop
               Neighbour := Translate (Point     => (X, Y),
                                       Direction => Direction);

               if Neighbour = (X, Y) then
                  null;
               elsif Item (Neighbour.X,
                           Neighbour.Y).Interpolation = Ignore then
                  null;
               else
                  Sum := Sum + Item (Neighbour.X, 
                                     Neighbour.Y).Height;
                  Count := Count + 1;
               end if;
            end loop;

            if Count > 0 then
               Item (X, Y) :=
                 (Interpolation => Flexible,
                  Height        => (Sum + Item (X, Y).Height * Midpoint_Weight)
                                     / (Float (Count) + Midpoint_Weight));
            end if;
         end if;
      end if;
   end Adjust;

   ---------------------------------------------------------------------------
   --  procedure Interpolate:

   procedure Interpolate (Item       : in out Landscape;
                          Iterations : in     Positive) is

   begin --  Interpolate
      for Iteration in 1 .. (Iterations + 3) / 4 loop
         for X in Item'Range (1) loop
            for Y in Item'Range (2) loop
               Adjust (Item => Item,
                       X    => X,
                       Y    => Y);
            end loop;
         end loop;

         for Y in Item'Range (2) loop
            for X in Item'Range (1) loop
               Adjust (Item => Item,
                       X    => X,
                       Y    => Y);
            end loop;
         end loop;

         for X in reverse Item'Range (1) loop
            for Y in Item'Range (2) loop
               Adjust (Item => Item,
                       X    => X,
                       Y    => Y);
            end loop;
         end loop;

         for Y in reverse Item'Range (2) loop
            for X in Item'Range (1) loop
               Adjust (Item => Item,
                       X    => X,
                       Y    => Y);
            end loop;
         end loop;
      end loop;
   end Interpolate;

   ---------------------------------------------------------------------------
   --  procedure Save:

   procedure Save (Item : in     Landscape;
                   Name : in     String) is
      use Ada.Integer_Text_IO;
      use Ada.Text_IO;

      Resolution : constant Positive := 255;
      Min   : Float := Float'Last;
      Max   : Float := Float'First;
      Image : File_Type;
   begin --  Save
      for X in Item'Range (1) loop
         for Y in Item'Range (2) loop
            if Item (X, Y).Interpolation = Ignore then
               null;
            else
               Min := Float'Min (Min, Item (X, Y).Height);
               Max := Float'Max (Max, Item (X, Y).Height);
            end if;
         end loop;
      end loop;

      Create (File => Image,
              Name => Name,
              Mode => Out_File);
      Put_Line (File => Image,
                Item => "P3");
      Put (File => Image,
           Item => Item'Length (1));
      Put (File => Image,
           Item => " ");
      Put (File => Image,
           Item => Item'Length (2));
      New_Line (File => Image);
      Put (File => Image,
           Item => Resolution);
      New_Line (File => Image);
      for Y in reverse Item'Range (2) loop
         for X in Item'Range (1) loop
            case Item (X, Y).Interpolation is
               when Ignore =>
                  Put (File => Image,
                       Item => Resolution);
                  Put (File => Image,
                       Item => " ");
                  Put (File => Image,
                       Item => 0);
                  Put (File => Image,
                       Item => " ");
                  Put (File => Image,
                       Item => 0);
               when Fixed =>
                  Put (File => Image,
                       Item => 127);
                  Put (File => Image,
                       Item => " ");
                  Put (File => Image,
                       Item => 0);
                  Put (File => Image,
                       Item => " ");
                  Put (File => Image,
                       Item => Natural (Float (Resolution) *
                                          (Item (X, Y).Height - Min) /
                                          (Max-Min)));
               when Flexible =>
                  Put (File => Image,
                       Item => 0);
                  Put (File => Image,
                       Item => " ");
                  Put (File => Image,
                       Item => Natural (Float (Resolution) *
                                          (Item (X, Y).Height - Min) /
                                          (Max-Min)));
                  Put (File => Image,
                       Item => " ");
                  Put (File => Image,
                       Item => 0);
            end case;

            if Col (File => Image) > 60 then
               New_Line (File => Image);
            else
               Put (File => Image,
                    Item => " ");
            end if;
         end loop;
      end loop;
      Close (File => Image);
   exception
      when others =>
         Put_Line (File => Current_Error,
                   Item => "Could not save landscape as """ & Name & """.");
   end Save;

   ---------------------------------------------------------------------------

end Landscapes;
