------------------------------------------------------------------------------
--
--  package Server_Communications (body)
--
------------------------------------------------------------------------------
--  Standard packages:

with Ada.Command_Line;
with Ada.Integer_Text_IO;
with Ada.Strings;
with Ada.Strings.Fixed;
with Ada.Strings.Unbounded;
with Ada.Text_IO;

------------------------------------------------------------------------------
--  Other packages:

with Sockets;

------------------------------------------------------------------------------

package body Server_Communications is

   ---------------------------------------------------------------------------
   --  Server connection:

   Server   : Sockets.Socket_FD;
   New_Line : constant Character := Character'Val (10);

   ---------------------------------------------------------------------------
   --  function Server_Name:

   function Server_Name return String is

      use Ada.Command_Line;
      use Ada.Text_IO;

      Default : constant String := "localhost";

   begin --  Server_Name
      if Argument_Count = 2 then
         return Argument (1);
      else
         Put_Line (File => Standard_Error,
                   Item => "Expected the server name as the first command " &
                           "line argument and the port number as the " &
                           "second command line argument. Using """ &
                           Default & """ as the server name.");
         return Default;
      end if;
   end Server_Name;

   ---------------------------------------------------------------------------
   --  function Server_Port_Number:

   function Server_Port_Number return Positive is

      use Ada.Command_Line;
      use Ada.Text_IO;

      Default : constant Positive := 2002;

   begin --  Server_Port_Number
      if Argument_Count = 2 then
         return Positive'Value (Argument (2));
      else
         Put_Line (File => Standard_Error,
                   Item => "Expected the server name as the first command " &
                           "line argument and the port number as the " &
                           "second command line argument. Using """ &
                           Positive'Image (Default) & """ as the server " &
                           "port number.");
         return Default;
      end if;
   exception
      when others =>
         Put_Line (File => Standard_Error,
                   Item => "The second command line argument was not a " &
                           "positive integer. Using """ &
                           Positive'Image (Default) & """ as the server " &
                           "port number.");
         return Default;
   end Server_Port_Number;

   ---------------------------------------------------------------------------
   --  Initialisation:

   procedure Initialise (Width  :    out Base_Types.X_Coordinate;
                         Height :    out Base_Types.Y_Coordinate) is
      use Ada.Text_IO;
      use Sockets;

      Initialisation_String : constant String := "Player";
   begin --  Initialise
      Socket (Sock => Server);
      Connect (Socket => Server,
               Host   => Server_Name,
               Port   => Server_Port_Number);
      Put (Socket => Server,
           Str    => Initialisation_String & New_Line);
      Put_Line (Item => "Sent....: """ & Initialisation_String & """");

  Get_Board_Dimensions:
      declare
         use Base_Types.X_Coordinate_Text_IO;
         use Base_Types.Y_Coordinate_Text_IO;

         Line : constant String := Get_Line (Socket => Server);
         Last : Natural := Line'First - 1;
      begin
         Put_Line (Item => "Received: """ & Line & """");

         Get (From => Line (Last + 1 .. Line'Last),
              Item => Width,
              Last => Last);
         Get (From => Line (Last + 1 .. Line'Last),
              Item => Height,
              Last => Last);
      end Get_Board_Dimensions;
   end Initialise;

   ---------------------------------------------------------------------------
   --  procedure Get_Board:

   procedure Get_Board (Width  : in     Base_Types.X_Coordinate;
                        Height : in     Base_Types.Y_Coordinate;
                        Board  :    out Base_Types.Board_Type) is

      use Ada.Text_IO;
      use Base_Types;
      use Sockets;

      Line       : String (1 .. Natural (Width));
      Coded_Tile : Character;

   begin --  Get_Board
      for X in Board'Range (1) loop
         for Y in Board'Range (2) loop
            Board (X, Y) := (Occupied            => False,
                             Tile                => Wall,
                             Net_Dropped_Gifts   => 0);
         end loop;
      end loop;

      for Y in 1 .. Height loop
         Line := Get_Line (Socket => Server);
         Put_Line (Item => "Received: """ & Line & """");

         for X in 1 .. Width loop
            Coded_Tile := Line (Natural (X));

            case Coded_Tile is
               when '.' => Board (X, Y) := (Occupied            => False,
                                            Tile                => Plain,
                                            Net_Dropped_Gifts   => 0);
               when '~' => Board (X, Y) := (Occupied            => False,
                                            Tile                => Water,
                                            Net_Dropped_Gifts   => 0);
               when '#' => Board (X, Y) := (Occupied            => False,
                                            Tile                => Wall,
                                            Net_Dropped_Gifts   => 0);
               when '@' => Board (X, Y) := (Occupied            => False,
                                            Tile                => Home,
                                            Net_Dropped_Gifts   => 0);
               when others =>
                  Put_Line (File => Standard_Error,
                            Item => "Invalid tile code (" &
                            Natural'Image (Character'Pos (Coded_Tile)) &
                            ") at (" & X'Img & "," & Y'Img & ").");
            end case;
         end loop;
      end loop;
   end Get_Board;

   ---------------------------------------------------------------------------

   procedure Get_Robot_Configuration
     (Robot :    out Base_Types.Robot_configuration) is

      use Ada.Text_IO;
      use Sockets;
      use Base_Types.Robot_ID_Text_IO;
      use Base_Types.Weight_Text_IO;
      use Base_Types.Money_Text_IO;

      Line : constant String := Get_Line (Socket => Server);
      Last : Natural := Line'First - 1;
   begin --  Get_Robot_Configuration
      Put_Line (Item => "Received: """ & Line & """");

      Get (From => Line (Last + 1 .. Line'Last),
           Item => Robot.ID,
           Last => Last);
      Get (From => Line (Last + 1 .. Line'Last),
           Item => Robot.Capacity,
           Last => Last);
      Get (From => Line (Last + 1 .. Line'Last),
           Item => Robot.Cash,
           Last => Last);
   end Get_Robot_Configuration;

   ---------------------------------------------------------------------------
   --  Receive and process server response:

   Action_String : Ada.Strings.Unbounded.Unbounded_String;
   Current_Robot : Base_Types.Robot_ID;

   ---------------------------------------------------------------------------
   --  procedure Receive_Action_Results:

   procedure Receive_Action_Results is
      use Ada.Strings;
      use Ada.Strings.Unbounded;
      use Ada.Text_IO;
      use Sockets;
   begin --  Receive_Action_Results
      Action_String := To_Unbounded_String (Get_Line (Server));
      Put_Line (Item => "Received: """ & To_String (Action_String) & """");
      Trim (Source => Action_String,
            Side   => Both);
   end Receive_Action_Results;

   ---------------------------------------------------------------------------
   --  function More_Action_Results:

   function More_Action_Results return Boolean is
      use Ada.Strings;
      use Ada.Strings.Unbounded;
--    use Base_Types.Robot_ID_Text_IO;

--    Last : Natural;
   begin --  More_Action_Results
--    loop
         Trim (Source => Action_String,
               Side   => Left);

--       exit when Length (Action_String) = 0;
--       exit when Element (Action_String, 1) /= '#';
--
--       Delete (Source  => Action_String,
--               From    => 1,
--               Through => 1);
--       Get (From => To_String (Action_String),
--            Item => Current_Robot,
--            Last => Last);
--       Delete (Source  => Action_String,
--               From    => 1,
--               Through => Last);
--    end loop;

      return Length (Action_String) > 0;
   end More_Action_Results;

   ---------------------------------------------------------------------------

   function Action_Robot_Change return Boolean is
      use Ada.Strings.Unbounded;
   begin
      return Element (Action_String, 1) = '#';
   end Action_Robot_Change;

   ---------------------------------------------------------------------------

   procedure Get_Action_Robot_Change (Robot :    out Base_Types.Robot_ID) is
      use Ada.Strings;
      use Ada.Strings.Unbounded;
      use Base_Types.Robot_ID_Text_IO;

      Last : Natural;
   begin --  More_Action_Results
      if Element (Action_String, 1) = '#' then
         Delete (Source  => Action_String,
                 From    => 1,
                 Through => 1);
         Get (From => To_String (Action_String),
              Item => Current_Robot,
              Last => Last);
         Delete (Source  => Action_String,
                 From    => 1,
                 Through => Last);
         Trim (Source => Action_String,
               Side   => Left);

         Robot := Current_Robot;
      else
         raise Constraint_Error;
      end if;
   end Get_Action_Robot_Change;

   ---------------------------------------------------------------------------

   procedure Get_Action_Result (Robot :    out Base_Types.Robot_ID;
                                Item  :    out Action_Result) is
      use Ada.Strings;
      use Ada.Strings.Unbounded;
      use Ada.Text_IO;
      use Base_Types;
      use Base_Types.Gift_ID_Text_IO;
      use Base_Types.X_Coordinate_Text_IO;
      use Base_Types.Y_Coordinate_Text_IO;

      Last : Natural;
      Key  : Character;
      Gift : Gift_ID;
      X    : X_Coordinate;
      Y    : Y_Coordinate;
   begin --  Get_Action_Result
      Robot := Current_Robot;

      Key := Element (Action_String, 1);
      Delete (Source  => Action_String,
              From    => 1,
              Through => 1);

      case Key is
         when 'N' =>
            Item := (Action => Moved_North);
         when 'E' =>
            Item := (Action => Moved_East);
         when 'S' =>
            Item := (Action => Moved_South);
         when 'W' =>
            Item := (Action => Moved_West);
         when 'P' =>
            Get (From => To_String (Action_String),
                 Item => Gift,
                 Last => Last);
            Delete (Source  => Action_String,
                    From    => 1,
                    Through => Last);

            Item := (Action => Picked_Gift_Up,
                     Gift   => Gift);
         when 'D' =>
            Get (From => To_String (Action_String),
                 Item => Gift,
                 Last => Last);
            Delete (Source  => Action_String,
                    From    => 1,
                    Through => Last);

            Item := (Action => Dropped_Gift,
                     Gift   => Gift);
         when 'X' =>
            Get (From => To_String (Action_String),
                 Item => X,
                 Last => Last);
            Delete (Source  => Action_String,
                    From    => 1,
                    Through => Last);
            Trim (Source => Action_String,
                  Side   => Left);

            Key := Element (Action_String, 1);
            Delete (Source  => Action_String,
                    From    => 1,
                    Through => 1);
            Trim (Source => Action_String,
                  Side   => Left);

            Get (From => To_String (Action_String),
                 Item => Y,
                 Last => Last);
            Delete (Source  => Action_String,
                    From    => 1,
                    Through => Last);

            Item := (Action => Set_XY,
                     X      => X,
                     Y      => Y);
         when others =>
            Put_Line (File => Standard_Error,
                      Item => "The server pretended that a robot performed " &
                              "the unknown action , " & Key & ".");
      end case;
   end Get_Action_Result;

   ---------------------------------------------------------------------------
   --  Available gifts:

   Gift_String : Ada.Strings.Unbounded.Unbounded_String;

   ---------------------------------------------------------------------------
   --  procedure Receive_Available_Gifts:

   procedure Receive_Available_Gifts is
      use Ada.Strings;
      use Ada.Strings.Unbounded;
      use Ada.Text_IO;
      use Sockets;
   begin --  Receive_Available_Gifts
      Gift_String := To_Unbounded_String (Get_Line (Server));
      Put_Line (Item => "Received: """ & To_String (Gift_String) & """");
      Trim (Source => Gift_String,
            Side   => Both);
   end Receive_Available_Gifts;

   ---------------------------------------------------------------------------

   function More_Available_Gifts return Boolean is
      use Ada.Strings;
      use Ada.Strings.Unbounded;
   begin --  More_Available_Gifts
      Trim (Source => Gift_String,
            Side   => Both);
      return Length (Gift_String) > 0;
   end More_Available_Gifts;

   ---------------------------------------------------------------------------

   procedure Get_Gift_Data (ID   :    out Base_Types.Gift_ID;
                            Data :    out Base_Types.Gift_Type) is
      use Ada.Strings.Unbounded;
      use Base_Types;
      use Base_Types.Gift_ID_Text_IO;
      use Base_Types.X_Coordinate_Text_IO;
      use Base_Types.Y_Coordinate_Text_IO;
      use Base_Types.Weight_Text_IO;

      Last : Natural;

      Destination : Position_Type;
      Weight      : Weight_Type;

   begin --  Get_Gift_Data
      Get (From => To_String (Gift_String),
           Item => ID,
           Last => Last);
      Delete (Source  => Gift_String,
              From    => 1,
              Through => Last);

      Get (From => To_String (Gift_String),
           Item => Destination.X,
           Last => Last);
      Delete (Source  => Gift_String,
              From    => 1,
              Through => Last);

      Get (From => To_String (Gift_String),
           Item => Destination.Y,
           Last => Last);
      Delete (Source  => Gift_String,
              From    => 1,
              Through => Last);

      Get (From => To_String (Gift_String),
           Item => Weight,
           Last => Last);
      Delete (Source  => Gift_String,
              From    => 1,
              Through => Last);

      Data := (Known       => True,
               ID          => ID,
               Destination => Destination,
               Weight      => Weight);
   end Get_Gift_Data;

   ------------------------------------------------------------------------
   --  Request action:

   procedure Request_Action (Bid  : in     Base_Types.Bid_Type;
                             Item : in     Base_Types.Command) is
      use Ada.Strings;
      use Ada.Strings.Fixed;
      use Ada.Text_IO;
      use Base_Types;
      use Sockets;

      Direction_Codes : array (Directions) of Character := (North => 'N',
                                                            South => 'S',
                                                            East  => 'E',
                                                            West  => 'W');
   begin --  Request_Action
      Put (Socket => Server,
           Str    => Trim (Source => Bid_Type'Image (Bid),
                           Side   => Both) & " ");
      Put (Item => "Sent....: """ & Trim (Source => Bid_Type'Image (Bid),
                                          Side   => Both) & " ");

      case Item.Action is
         when Move =>
            Put (Socket => Server,
                 Str    => "Move " & Direction_Codes (Item.Direction));
            Put (Item   => "Move " & Direction_Codes (Item.Direction));
         when Pick =>
            Put (Socket => Server,
                 Str    => "Pick");
            Put (Item   => "Pick");
            for Index in Item.Gifts'Range loop
               Put (Socket => Server,
                    Str    => " " & Gift_ID'Image (Item.Gifts (Index)));
               Put (Item   => " " & Gift_ID'Image (Item.Gifts (Index)));
            end loop;
         when Drop =>
            Put (Socket => Server,
                 Str    => "Drop");
            Put (Item   => "Drop");
            for Index in Item.Gifts'Range loop
               Put (Socket => Server,
                    Str    => " " & Gift_ID'Image (Item.Gifts (Index)));
               Put (Item   => " " & Gift_ID'Image (Item.Gifts (Index)));
            end loop;
      end case;

      Put (Socket => Server,
           Str    => (1 => New_Line));
      Put_Line (Item => """");
   end Request_Action;

   ------------------------------------------------------------------------

end Server_Communications;
